#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Pieces of any ingestion manifest """

import abc
import re
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Union

from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import AncillaryProductType, ScienceProductType, Telescope

MANIFEST_FILENAME = "ingestion-manifest.json"
CURATOR_MANIFEST_FILENAME = "manifest.json"
INGESTION_ARTIFACTS_NAME = "ingestion_artifacts_"
TARFILE_EXT = ".tar"
WEBLOG_FILENAME = "weblog.tgz"
INIT_WEBLOG_FILENAME = "initial_weblog.tgz"
REIMAGING_FILENAME = "reimaging_resources.tgz"
PARAMETER_FILENAME = "SEIP_parameter.list"
SELFCAL_AUX_FILENAME = "selfcal.auxproducts.tgz"
SCIENCE_PRODUCT_PATTERN = re.compile("[a-zA-Z0-9._\\-+]*\\.tar")
JSON = Union[int, float, str, List["JSON"], Dict[str, "JSON"]]


class IngestionManifestKey(Enum):
    """Sections we expect to see in a manifest"""

    PARAMETERS = "parameters"
    INPUT_GROUP = "input_group"
    OUTPUT_GROUP = "output_group"
    SCIENCE_PRODUCTS = "science_products"
    ANCILLARY_PRODUCTS = "ancillary_products"
    REINGEST = "reingest"


class ParamsKey(Enum):
    """Keys in ManifestParameters"""

    PARAMETERS = "parameters"
    TELESCOPE = "telescope"
    REINGEST = "reingest"
    NGAS_INGEST = "ngas_ingest"
    INGESTION_PATH = "ingestion_path"
    ADDITIONAL_METADATA = "additional_metadata"
    COLLECTION_METADATA = "additional_metadata"  # needed for VLASS, realfast, elwa, alfalfa. Not yet implemented


class ReingestKey(Enum):
    """Keys in ReingestGroup"""

    TARGETS = "targets"
    LOCATOR = "locator"
    TYPE = "type"
    PRODUCT_GROUP = "product_group"


class ManifestComponentIF(abc.ABC):
    """Interface for components of an ingestion manifest (including the manifest itself)"""

    @abc.abstractmethod
    def __eq__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def to_dict(self) -> Dict:
        """
        Build a JSON-ifyable dictionary

        :return: json.dumps()-able dictionary object
        """


@dataclass
class InputScienceProduct(ManifestComponentIF):
    """Simplest type of science product: has only a locator"""

    locator: str

    def to_dict(self) -> JSON:
        """
        Turn me into a json-ifiable dict

        :return: dicty-me
        """
        return {"locator": self.locator}


@dataclass
class InputGroup(ManifestComponentIF):
    """Generic ingestion manifest input group"""

    science_products: List[InputScienceProduct]

    def to_dict(self) -> JSON:
        """
        Turn me into a json-ifiable dict

        :return: dicty-me
        """
        me_dict = dict(self.__dict__)

        sps = me_dict[IngestionManifestKey.SCIENCE_PRODUCTS.value]
        sps = [sp.to_dict() for sp in sps]
        if len(sps) == 0:
            return {}

        to_return = {IngestionManifestKey.SCIENCE_PRODUCTS.value: sps}
        return to_return


@dataclass
class ManifestParameters(ManifestComponentIF):
    """Represents "parameters" section of ingestion manifest.
    ASSUMPTIONS:
    * "ngas_ingest" is always True (per our testing examples)
    """

    telescope: Telescope
    ngas_ingest: bool
    staging_source_dir: Optional[Path]
    additional_metadata: Optional[AbstractTextFile] = None

    def to_dict(self) -> JSON:
        """
        form dictionary object for this manifest section

        :return:
        """
        json_dict = {
            ParamsKey.TELESCOPE.value: self.telescope.value,
            # The ingestion manifest must have "true" and "false"
            # rather than "True" and "False"
            ParamsKey.NGAS_INGEST.value: self.ngas_ingest,
        }
        if self.staging_source_dir is not None:
            json_dict[ParamsKey.INGESTION_PATH.value] = str(self.staging_source_dir)
        if self.additional_metadata:
            json_dict[ParamsKey.ADDITIONAL_METADATA.value] = self.additional_metadata.filename

        return json_dict


@dataclass
class AncillaryProduct(ManifestComponentIF):
    """Represents an ancillary product in an ingestion manifest"""

    type: AncillaryProductType
    filename: str
    science_associate: Optional[str] = None
    group_with: Optional[str] = None

    def __str__(self):
        return f"{self.filename}: {self.type.value}"

    def to_dict(self) -> JSON:
        """
        Turn me into a json-ifiable dict

        :return: dicty-me
        """
        me_dict = dict(self.__dict__)
        me_dict["type"] = self.type.value
        clean_dict = {}
        for key, val in me_dict.items():
            if val is not None:
                clean_dict[key] = val

        return clean_dict


@dataclass
class OutputScienceProduct(ManifestComponentIF):
    """Generic science product contained in manifest output group"""

    type: Union[ScienceProductType, AncillaryProductType]
    filename: str
    ancillary_products: Optional[List[AncillaryProduct]] = None

    def __str__(self):
        return f"{Path(self.filename).name}: {self.type.value}, {len(self.ancillary_products)} ancillary products"

    def to_dict(self) -> JSON:
        json_dict = {"type": self.type.value, "filename": self.filename}
        if self.ancillary_products:
            aps_json = [ap.to_dict() for ap in self.ancillary_products]
            json_dict[IngestionManifestKey.ANCILLARY_PRODUCTS.value] = aps_json
        return json_dict


@dataclass
class OutputGroup(ManifestComponentIF):
    """Generic ingestion manifest output group"""

    science_products: List[OutputScienceProduct]
    ancillary_products: Optional[List[AncillaryProduct]] = None

    def to_dict(self) -> JSON:
        """
        Turn me into a json-ifiable dict

        :return: dicty-me
        """
        me_dict = dict(self.__dict__)

        sps = me_dict[IngestionManifestKey.SCIENCE_PRODUCTS.value]
        sps = [sp.to_dict() for sp in sps]
        me_dict[IngestionManifestKey.SCIENCE_PRODUCTS.value] = sps

        aps = me_dict[IngestionManifestKey.ANCILLARY_PRODUCTS.value]
        if aps:
            ap_jsons = [ap.to_dict() for ap in aps]
            me_dict[IngestionManifestKey.ANCILLARY_PRODUCTS.value] = ap_jsons

        return me_dict


@dataclass
class ReingestGroup(ManifestComponentIF):
    """Generic manifest reingest group"""

    product_type: ScienceProductType
    locator: Optional[str] = None  # Replaced by product_group for full curation
    product_group: Optional[int] = None  # Replaced by locator for partial curation
    targets: Optional[List[str]] = None  # optional: presence determines Partial vs Full curation.

    def to_dict(self) -> JSON:
        """
        Turn me into a json-ifiable dict, excluding None fields

        :return: dict
        """
        key_value_pairs = [
            (ReingestKey.TARGETS.value, self.targets),
            (ReingestKey.LOCATOR.value, self.locator),
            (ReingestKey.TYPE.value, self.product_type.value),
            (ReingestKey.PRODUCT_GROUP.value, self.product_group),
        ]
        json_dict = dict()
        for key, value in key_value_pairs:
            if value is not None:
                json_dict[key] = value

        return json_dict
